function Api() {
    this._requests = [];
    this._lang = "ru";
    this._csrf_token = "token";
    this._base_url = "https://tradeback.io";
}

Api.prototype.request = function (url, options, callback) {
    // для повторного запроса при обновлении csrf
    let clean_args = [url, options, callback];

    if (typeof options === 'function') {
        callback = options;
        options = undefined;
    }
    if (Object(url) === url) {
        options = url;
        url = undefined;
    }

    // клонируем options
    options = Object.assign({}, options);

    let self = this,
        headers = options.headers,
        completeFunc = options.complete,
        isApiCall = false;

    // удаляем параметры которые требует ручной обработки
    delete options.headers;
    delete options.complete;

    // если url строка, то тогда она должна перезаписывать url из options
    if (typeof url === 'string') {
        options.url = url;
    }

    // валидируем options.url
    if (typeof options.url !== 'string') {
        throw new Error('Url must be a string');
    }

    if (options.prefix !== false) {
        // добавляем /api/ в ссылке если отсутствует
        options.url = options.url.replace(/^(\/?api\/?|\/?)/, '/api/');
    }

    // проверяем если это обращение к API
    if ((/^\/api\//).test(options.url)) {
        isApiCall = true;
    }

    if (isApiCall && !(/(\?|&)lang/).test(options.url)) {
        // устанавливаем язык в GET параметрах для запросах к API если еще не указан
        options.url += (options.url.indexOf('?') === -1 ? '?' : '&') + 'lang=' + self._lang;
    }

    if (!(/(http|https)/).test(options.url)) {
        // добавляем _base_url если домен не указан явно
        options.url = self._base_url + options.url;
    }

    // проверяем если метод запроса не GET или HEAD и если data объект для запросах к API
    if (
        isApiCall
        && ['GET', 'HEAD'].indexOf(String(options.method).toUpperCase()) === -1
        && Object(options.data) === options.data
    ) {
        options.data = JSON.stringify(options.data);
    }

    let req = $.ajax(Object.assign({
            method: 'POST',
            dataType: 'json',
            contentType: 'application/json',
            headers: Object.assign({
                'X-CSRF-TOKEN': self._csrf_token
            }, headers),
            complete: function (xhr, status) {
                // удаляем данный запрос из списка активных запросов
                for (let i = 0, l = self._requests.length; i < l; i++) {
                    if (self._requests[i] !== req) continue;

                    // удаляем запрос из массива
                    self._requests.splice(i, 1);

                    break;
                }

                if (xhr.status === 419) {
                    // не правильный csrf token
                    return self.updateCsrf(clean_args);
                }

                // вызываем колбэки указанные в options
                if (Array.isArray(completeFunc)) {
                    for (let i = 0, l = completeFunc.length; i < l; i++) {
                        callCallback(completeFunc[i], [xhr, status]);
                    }
                }
                else if (typeof completeFunc === 'function') {
                    callCallback(completeFunc, [xhr, status]);
                }

                let json, message,
                    callbackArgs = [null, xhr];


                if (Object(xhr.responseJSON) !== xhr.responseJSON) {
                    // ответ в формате Text
                    callbackArgs[2] = xhr.responseText;

                    callCallback(callback, callbackArgs);

                    return;
                } else {
                    // ответ в формате json
                    json = xhr.responseJSON;

                    callbackArgs[2] = json;

                    // проверяем статус для запросах к API и ответ от сервера если все успешно
                    if ((isApiCall && !json.success) || status !== 'success') {
                        message = isApiCall && json.message ? json.message : 'error.Unknown';

                        //callbackArgs[0] = new Error(message);
                        callbackArgs[0] = message;

                        callCallback(callback, callbackArgs);

                        return;
                    }

                    callCallback(callback, callbackArgs);

                }
            }
        }, options));

    // добавляем запрос в список активных запросов
    self._requests.push(req);

    function callCallback(func, args) {
        if (typeof func !== 'function') return;

        try {
            func.apply(null, args);
        } catch(e) {}
    }
};

// обновить csrf и повторить запрос в котором была вызвана ошибка
Api.prototype.updateCsrf = function ([url, options, callback]) {
    let self = this;

    self.request({
        method : 'GET',
        url : '/get/csrf-token',
        prefix : false
    }, function (err, res, body) {
        if (err) return console.error(err);

        if (Object(body) !== body) {
            try {
                body = JSON.parse(body);
            } catch (e) {}
        }

        if (!body.token) return console.error('Can\'t update CSRF');

        self._csrf_token = body.token;

        // повторим запрос в котором была ошибка
        return self.request(url, options, callback);
    });

}

var api = new Api();